"""Code is adapted from https://github.com/jerrywn121/TianChi_AIEarth/blob/main/SAConvLSTM/utils.py"""
import os
from typing import Optional
import numpy as np
from torch.utils.data import Dataset, DataLoader
import xarray as xr
from pathlib import Path
import torch
from pytorch_lightning import LightningDataModule
from ...config import cfg


NINO_WINDOW_T = 3  # Nino index is the sliding average over sst, window size is 3
CMIP6_SST_MAX = 10.198975563049316
CMIP6_SST_MIN = -16.549121856689453
CMIP5_SST_MAX = 8.991744995117188
CMIP5_SST_MIN = -9.33076286315918
CMIP6_NINO_MAX = 4.138188362121582
CMIP6_NINO_MIN = -3.5832221508026123
CMIP5_NINO_MAX = 3.8253555297851562
CMIP5_NINO_MIN = -2.691682815551758
SST_MAX = max(CMIP6_SST_MAX, CMIP5_SST_MAX)
SST_MIN = min(CMIP6_SST_MIN, CMIP5_SST_MIN)

default_enso_dir = os.path.join(cfg.datasets_dir, "icar_enso_2021")
# unzipped file is saved in `default_enso_nc_dir`.
default_enso_nc_dir = os.path.join(default_enso_dir, "enso_round1_train_20210201")

def scale_sst(sst):
    return (sst - SST_MIN) / (SST_MAX - SST_MIN)

def scale_back_sst(sst):
    return (SST_MAX - SST_MIN) * sst + SST_MIN

def prepare_inputs_targets(len_time, input_gap, input_length, pred_shift, pred_length, samples_gap):
    # input_gap=1: time gaps between two consecutive input frames
    # input_length=12: the number of input frames
    # pred_shift=26: the lead_time of the last target to be predicted
    # pred_length=26: the number of frames to be predicted
    assert pred_shift >= pred_length
    input_span = input_gap * (input_length - 1) + 1
    pred_gap = pred_shift // pred_length
    input_ind = np.arange(0, input_span, input_gap)
    target_ind = np.arange(0, pred_shift, pred_gap) + input_span + pred_gap - 1
    ind = np.concatenate([input_ind, target_ind]).reshape(1, input_length + pred_length)
    max_n_sample = len_time - (input_span + pred_shift - 1)
    ind = ind + np.arange(max_n_sample)[:, np.newaxis] @ np.ones((1, input_length + pred_length), dtype=int)
    return ind[::samples_gap]

def fold(data, size=36, stride=12):
    # inverse of unfold/sliding window operation
    # only applicable to the case where the size of the sliding windows is n*stride
    # data (N, size, *)
    # outdata (N_, *)
    # N/size is the number/width of sliding blocks
    assert size % stride == 0
    times = size // stride
    remain = (data.shape[0] - 1) % times
    if remain > 0:
        ls = list(data[::times]) + [data[-1, -(remain * stride):]]
        outdata = np.concatenate(ls, axis=0)  # (36*(151//3+1)+remain*stride, *, 15)
    else:
        outdata = np.concatenate(data[::times], axis=0)  # (36*(151/3+1), *, 15)
    assert outdata.shape[0] == size * ((data.shape[0] - 1) // times + 1) + remain * stride
    return outdata

def data_transform(data, num_years_per_model):
    # data (N, 36, *)
    # num_years_per_model: 151/140
    length = data.shape[0]
    assert length % num_years_per_model == 0
    num_models = length // num_years_per_model
    outdata = np.stack(np.split(data, length / num_years_per_model, axis=0), axis=-1)  # (151, 36, *, 15)
    # cmip6sst outdata.shape = (151, 36, 24, 48, 15) = (year, month, lat, lon, model)
    # cmip5sst outdata.shape = (140, 36, 24, 48, 17)
    # cmip6nino outdata.shape = (151, 36, 15)
    # cmip5nino outdata.shape = (140, 36, 17)
    outdata = fold(outdata, size=36, stride=12)
    # cmip6sst outdata.shape = (1836, 24, 48, 15), 1836 == 151 * 12 + 24
    # cmip5sst outdata.shape = (1704, 24, 48, 17)
    # cmip6nino outdata.shape = (1836, 15)
    # cmip5nino outdata.shape = (1704, 17)

    # check output data
    assert outdata.shape[-1] == num_models
    assert not np.any(np.isnan(outdata))
    return outdata

def read_raw_data(ds_dir, out_dir=None):
    # read and process raw cmip data from CMIP_train.nc and CMIP_label.nc
    train_cmip = xr.open_dataset(Path(ds_dir) / 'CMIP_train.nc').transpose('year', 'month', 'lat', 'lon')
    label_cmip = xr.open_dataset(Path(ds_dir) / 'CMIP_label.nc').transpose('year', 'month')
    # train_cmip.sst.values.shape = (4645, 36, 24, 48)

    # select longitudes
    lon = train_cmip.lon.values
    lon = lon[np.logical_and(lon >= 95, lon <= 330)]
    train_cmip = train_cmip.sel(lon=lon)

    cmip6sst = data_transform(data=train_cmip.sst.values[:2265],
                              num_years_per_model=151)
    cmip5sst = data_transform(data=train_cmip.sst.values[2265:],
                              num_years_per_model=140)
    cmip6nino = data_transform(data=label_cmip.nino.values[:2265],
                               num_years_per_model=151)
    cmip5nino = data_transform(data=label_cmip.nino.values[2265:],
                               num_years_per_model=140)
    # cmip6sst.shape = (1836, 24, 48, 15)
    # cmip5sst.shape = (1704, 24, 48, 17)
    assert len(cmip6sst.shape) == 4
    assert len(cmip5sst.shape) == 4
    assert len(cmip6nino.shape) == 2
    assert len(cmip5nino.shape) == 2
    # store processed data for faster data access
    if out_dir is not None:
        ds_cmip6 = xr.Dataset({'sst': (['month', 'lat', 'lon', 'model'], cmip6sst),
                               'nino': (['month', 'model'], cmip6nino)},
                              coords={'month': np.repeat(np.arange(1, 13)[None], cmip6nino.shape[0] // 12,
                                                         axis=0).flatten(),
                                      'lat': train_cmip.lat.values, 'lon': train_cmip.lon.values,
                                      'model': np.arange(15) + 1})
        ds_cmip6.to_netcdf(Path(out_dir) / 'cmip6.nc')
        ds_cmip5 = xr.Dataset({'sst': (['month', 'lat', 'lon', 'model'], cmip5sst),
                               'nino': (['month', 'model'], cmip5nino)},
                              coords={'month': np.repeat(np.arange(1, 13)[None], cmip5nino.shape[0] // 12,
                                                         axis=0).flatten(),
                                      'lat': train_cmip.lat.values, 'lon': train_cmip.lon.values,
                                      'model': np.arange(17) + 1})
        ds_cmip5.to_netcdf(Path(out_dir) / 'cmip5.nc')
    train_cmip.close()
    label_cmip.close()
    return cmip6sst, cmip5sst, cmip6nino, cmip5nino

def read_from_nc(ds_dir):
    # an alternative for reading processed data
    cmip6 = xr.open_dataset(Path(ds_dir) / 'cmip6.nc').transpose('month', 'lat', 'lon', 'model')
    cmip5 = xr.open_dataset(Path(ds_dir) / 'cmip5.nc').transpose('month', 'lat', 'lon', 'model')
    return cmip6.sst.values, cmip5.sst.values, cmip6.nino.values, cmip5.nino.values

def cat_over_last_dim(data):
    r"""
    treat different models (15 from CMIP6, 17 from CMIP5) as batch_size
    e.g., cmip6sst.shape = (178, 38, 24, 48, 15), converted_cmip6sst.shape = (2670, 38, 24, 48)
    e.g., cmip5sst.shape = (165, 38, 24, 48, 15), converted_cmip6sst.shape = (2475, 38, 24, 48)
    """
    return np.concatenate(np.moveaxis(data, -1, 0), axis=0)

class cmip_dataset(Dataset):

    def __init__(self,
                 sst_cmip6, nino_cmip6,
                 sst_cmip5, nino_cmip5,
                 samples_gap,
                 in_len=12,
                 out_len=26,
                 in_stride=1,
                 out_stride=1,
                 normalize_sst=True):
        r"""
        Parameters
        ----------
        sst_cmip6
        nino_cmip6
        sst_cmip5
        nino_cmip5
        samples_gap: int
            stride of seq sampling.
            e.g., samples_gap = 10, the first seq contains [0, 1, ..., T-1] frame indices, the second seq contains [10, 11, .., T+9]
        """
        super().__init__()
        self.normalize_sst = normalize_sst
        # cmip6 (N, *, 15)
        # cmip5 (N, *, 17)
        sst = []
        target_nino = []

        nino_idx_slice = slice(in_len, in_len + out_len - NINO_WINDOW_T + 1)  # e.g., 12:36
        if sst_cmip6 is not None:
            assert len(sst_cmip6.shape) == 4
            assert len(nino_cmip6.shape) == 2
            idx_sst = prepare_inputs_targets(
                len_time=sst_cmip6.shape[0],
                input_length=in_len, input_gap=in_stride,
                pred_shift=out_len * out_stride, pred_length=out_len,
                samples_gap=samples_gap)

            sst.append(cat_over_last_dim(sst_cmip6[idx_sst]))
            target_nino.append(cat_over_last_dim(nino_cmip6[idx_sst[:, nino_idx_slice]]))
        if sst_cmip5 is not None:
            assert len(sst_cmip5.shape) == 4
            assert len(nino_cmip5.shape) == 2
            idx_sst = prepare_inputs_targets(
                len_time=sst_cmip5.shape[0],
                input_length=in_len, input_gap=in_stride,
                pred_shift=out_len * out_stride, pred_length=out_len,
                samples_gap=samples_gap)
            sst.append(cat_over_last_dim(sst_cmip5[idx_sst]))
            target_nino.append(cat_over_last_dim(nino_cmip5[idx_sst[:, nino_idx_slice]]))

        # sst data containing both the input and target
        self.sst = np.concatenate(sst, axis=0)  # (N, in_len+out_len, lat, lon)
        if normalize_sst:
            self.sst = scale_sst(self.sst)
        # nino data containing the target only
        self.target_nino = np.concatenate(target_nino, axis=0)  # (N, out_len+NINO_WINDOW_T-1)
        assert self.sst.shape[0] == self.target_nino.shape[0]
        assert self.sst.shape[1] == in_len + out_len
        assert self.target_nino.shape[1] == out_len - NINO_WINDOW_T + 1

    def GetDataShape(self):
        return {'sst': self.sst.shape,
                'nino target': self.target_nino.shape}

    def __len__(self):
        return self.sst.shape[0]

    def __getitem__(self, idx):
        return self.sst[idx], self.target_nino[idx]

class ENSOLightningDataModule(LightningDataModule):

    def __init__(self,
                 data_dir=default_enso_nc_dir,
                 in_len=12,
                 out_len=26,
                 in_stride=1,
                 out_stride=1,
                 train_samples_gap=10,
                 eval_samples_gap=11,
                 normalize_sst=True,
                 # datamodule_only
                 batch_size=1,
                 num_workers=8, ):
        super(ENSOLightningDataModule, self).__init__()
        self.data_dir = data_dir

        self.in_len = in_len
        self.out_len = out_len
        self.in_stride = in_stride
        self.out_stride = out_stride
        self.train_samples_gap = train_samples_gap
        self.eval_samples_gap = eval_samples_gap
        self.normalize_sst = normalize_sst
        # datamodule_only
        self.batch_size = batch_size
        assert num_workers == 1, ValueError(f"Current implementation does not support `num_workers != 1`!")
        self.num_workers = num_workers

    def prepare_data(self):
        if not os.path.exists(self.data_dir):
            raise ValueError(
                f"ENSO data_dir {self.data_dir} not exits! Follow README.md to download and unzip dataset!")

    def setup(self, stage: Optional[str] = None):
        cmip6sst, cmip5sst, cmip6nino, cmip5nino = read_raw_data(self.data_dir)
        # TODO: more flexible train/val/test split
        sst_train = [cmip6sst, cmip5sst[..., :-2]]
        nino_train = [cmip6nino, cmip5nino[..., :-2]]
        sst_eval = [cmip5sst[..., -2:-1]]
        nino_eval = [cmip5nino[..., -2:-1]]
        sst_test = [cmip5sst[..., -1:]]
        nino_test = [cmip5nino[..., -1:]]

        if stage == "fit" or stage is None:
            self.enso_train = cmip_dataset(
                sst_cmip6=sst_train[0], nino_cmip6=nino_train[0],
                sst_cmip5=sst_train[1], nino_cmip5=nino_train[1],
                samples_gap=self.train_samples_gap,
                in_len=self.in_len,
                out_len=self.out_len,
                in_stride=self.in_stride,
                out_stride=self.out_stride,
                normalize_sst=self.normalize_sst)
            self.enso_val = cmip_dataset(
                sst_cmip6=None, nino_cmip6=None,
                sst_cmip5=sst_eval[0], nino_cmip5=nino_eval[0],
                samples_gap=self.eval_samples_gap,
                in_len=self.in_len,
                out_len=self.out_len,
                in_stride=self.in_stride,
                out_stride=self.out_stride,
                normalize_sst=self.normalize_sst)

        if stage == "test" or stage is None:
            self.enso_test = cmip_dataset(
                sst_cmip6=None, nino_cmip6=None,
                sst_cmip5=sst_test[0], nino_cmip5=nino_test[0],
                samples_gap=self.eval_samples_gap,
                in_len=self.in_len,
                out_len=self.out_len,
                in_stride=self.in_stride,
                out_stride=self.out_stride,
                normalize_sst=self.normalize_sst)

        if stage == "predict" or stage is None:
            self.enso_predict = cmip_dataset(
                sst_cmip6=None, nino_cmip6=None,
                sst_cmip5=sst_test[0], nino_cmip5=nino_test[0],
                samples_gap=self.eval_samples_gap,
                in_len=self.in_len,
                out_len=self.out_len,
                in_stride=self.in_stride,
                out_stride=self.out_stride,
                normalize_sst=self.normalize_sst)

    @property
    def num_train_samples(self):
        return len(self.enso_train)

    @property
    def num_val_samples(self):
        return len(self.enso_val)

    @property
    def num_test_samples(self):
        return len(self.enso_test)

    @property
    def num_predict_samples(self):
        return len(self.enso_predict)

    def train_dataloader(self):
        return DataLoader(self.enso_train,
                          shuffle=True,
                          batch_size=self.batch_size,
                          num_workers=self.num_workers)

    def val_dataloader(self):
        return DataLoader(self.enso_val,
                          shuffle=False,
                          batch_size=self.batch_size,
                          num_workers=self.num_workers)

    def test_dataloader(self):
        return DataLoader(self.enso_test,
                          shuffle=False,
                          batch_size=self.batch_size,
                          num_workers=self.num_workers)

    def predict_dataloader(self):
        return DataLoader(self.enso_predict,
                          shuffle=False,
                          batch_size=self.batch_size,
                          num_workers=self.num_workers)



import os
import argparse
from shutil import copyfile
from omegaconf import OmegaConf
import numpy as np
from earthformer.datasets.nbody.nbody_mnist_torch_wrap import NBodyMovingMNISTLightningDataModule
from earthformer.visualization.utils import save_gif


def duplicate_single_seq_dataset(npz_path, num_copies=64, save_path=None):
    r"""
    Duplicate a single N-body sequence `num_copies` times as a toy dataset for debugging.
    """
    if save_path is None:
        from pathlib import Path
        save_dir = os.path.dirname(npz_path)
        save_name = Path(npz_path).stem + f"_copy{num_copies}.npz"
        save_path = os.path.join(save_dir, save_name)
    f = np.load(npz_path)
    saved_data_dict = dict(f)
    new_data_dict = {}
    for key, val in saved_data_dict.items():
        dup_val = val[0]
        dup_val = np.repeat(dup_val[np.newaxis],
                            repeats=num_copies,
                            axis=0)
        new_data_dict[key] = dup_val
    np.savez_compressed(file=save_path, **new_data_dict)

def generate_nbody_dataset(save_dir=None, oc_file_path=None):
    if oc_file_path is None:
        oc_file_path = os.path.join(os.path.dirname(os.path.realpath(__file__)), "cfg.yaml")
    dataset_oc = OmegaConf.to_object(OmegaConf.load(open(oc_file_path, "r")).dataset)
    if save_dir is None:
        from earthformer.config import cfg
        save_dir = os.path.join(cfg.datasets_dir, "nbody")
    data_dir = os.path.join(save_dir, dataset_oc["dataset_name"])
    if os.path.exists(data_dir):
        raise ValueError(f"data_dir {data_dir} already exists!")
    else:
        os.makedirs(data_dir)
    copyfile(oc_file_path, os.path.join(data_dir, "nbody_dataset_cfg.yaml"))
    dm = NBodyMovingMNISTLightningDataModule(
        data_dir=data_dir,
        force_regenerate=False,
        num_train_samples=dataset_oc["num_train_samples"],
        num_val_samples=dataset_oc["num_val_samples"],
        num_test_samples=dataset_oc["num_test_samples"],
        digit_num=dataset_oc["digit_num"],
        img_size=dataset_oc["img_size"],
        raw_img_size=dataset_oc["raw_img_size"],
        seq_len=dataset_oc["seq_len"],
        raw_seq_len_multiplier=dataset_oc["raw_seq_len_multiplier"],
        distractor_num=dataset_oc["distractor_num"],
        distractor_size=dataset_oc["distractor_size"],
        max_velocity_scale=dataset_oc["max_velocity_scale"],
        initial_velocity_range=dataset_oc["initial_velocity_range"],
        random_acceleration_range=dataset_oc["random_acceleration_range"],
        scale_variation_range=dataset_oc["scale_variation_range"],
        rotation_angle_range=dataset_oc["rotation_angle_range"],
        illumination_factor_range=dataset_oc["illumination_factor_range"],
        period=dataset_oc["period"],
        global_rotation_prob=dataset_oc["global_rotation_prob"],
        index_range=dataset_oc["index_range"],
        mnist_data_path=dataset_oc["mnist_data_path"],
        # N-Body params
        nbody_acc_mode=dataset_oc["nbody_acc_mode"],
        nbody_G=dataset_oc["nbody_G"],
        nbody_softening_distance=dataset_oc["nbody_softening_distance"],
        nbody_mass=dataset_oc["nbody_mass"],
        # datamodule_only
        batch_size=1,
        num_workers=8, )
    dm.prepare_data()
    dm.setup()
    seq = dm.nbody_train[0]
    save_gif(single_seq=seq[:, :, :, 0].numpy().astype(np.float32),
             fname=os.path.join(data_dir, "nbody_example.gif"))

def get_parser():
    parser = argparse.ArgumentParser()
    parser.add_argument('--cfg', default=None, type=str)
    return parser

def main():
    parser = get_parser()
    args = parser.parse_args()
    generate_nbody_dataset(oc_file_path=args.cfg)

if __name__ == "__main__":
    main()



"""Main module."""

import os
import pkg_resources
from .common import *

def get_pkg_dir(name="earthformer"):
    pkg_dir = os.path.dirname(pkg_resources.resource_filename(name, name + ".py"))
    return pkg_dir


def get_earthnet_yaml():
    pkg_dir = get_pkg_dir()
    earthnet_yaml = os.path.join(
        pkg_dir,
        "scripts/cuboid_transformer/earthnet_w_meso/earthformer_earthnet_v1.yaml",
    )
    if os.path.exists(earthnet_yaml):
        return earthnet_yaml
    else:
        raise FileNotFoundError(f"File not found: {earthnet_yaml}")



import numpy as np
import torch
from torch import nn
from torch.nn import functional as F


def round_to(dat, c):
    return dat + (dat - dat % c) % c

def get_activation(act, inplace=False, **kwargs):
    """

    Parameters
    ----------
    act
        Name of the activation
    inplace
        Whether to perform inplace activation

    Returns
    -------
    activation_layer
        The activation
    """
    if act is None:
        return lambda x: x
    if isinstance(act, str):
        if act == 'leaky':
            negative_slope = kwargs.get("negative_slope", 0.1)
            return nn.LeakyReLU(negative_slope, inplace=inplace)
        elif act == 'identity':
            return nn.Identity()
        elif act == 'elu':
            return nn.ELU(inplace=inplace)
        elif act == 'gelu':
            return nn.GELU()
        elif act == 'relu':
            return nn.ReLU()
        elif act == 'sigmoid':
            return nn.Sigmoid()
        elif act == 'tanh':
            return nn.Tanh()
        elif act == 'softrelu' or act == 'softplus':
            return nn.Softplus()
        elif act == 'softsign':
            return nn.Softsign()
        else:
            raise NotImplementedError('act="{}" is not supported. '
                                      'Try to include it if you can find that in '
                                      'https://pytorch.org/docs/stable/nn.html'.format(act))
    else:
        return act

class RMSNorm(nn.Module):
    def __init__(self, d, p=-1., eps=1e-8, bias=False):
        """Root Mean Square Layer Normalization proposed in "[NeurIPS2019] Root Mean Square Layer Normalization"

        Parameters
        ----------
        d
            model size
        p
            partial RMSNorm, valid value [0, 1], default -1.0 (disabled)
        eps
            epsilon value, default 1e-8
        bias
            whether use bias term for RMSNorm, disabled by
            default because RMSNorm doesn't enforce re-centering invariance.
        """
        super(RMSNorm, self).__init__()

        self.eps = eps
        self.d = d
        self.p = p
        self.bias = bias

        self.scale = nn.Parameter(torch.ones(d))
        self.register_parameter("scale", self.scale)

        if self.bias:
            self.offset = nn.Parameter(torch.zeros(d))
            self.register_parameter("offset", self.offset)

    def forward(self, x):
        if self.p < 0. or self.p > 1.:
            norm_x = x.norm(2, dim=-1, keepdim=True)
            d_x = self.d
        else:
            partial_size = int(self.d * self.p)
            partial_x, _ = torch.split(x, [partial_size, self.d - partial_size], dim=-1)

            norm_x = partial_x.norm(2, dim=-1, keepdim=True)
            d_x = partial_size

        rms_x = norm_x * d_x ** (-1. / 2)
        x_normed = x / (rms_x + self.eps)

        if self.bias:
            return self.scale * x_normed + self.offset

        return self.scale * x_normed

def get_norm_layer(normalization: str = 'layer_norm',
                   axis: int = -1,
                   epsilon: float = 1e-5,
                   in_channels: int = 0, **kwargs):
    """Get the normalization layer based on the provided type

    Parameters
    ----------
    normalization
        The type of the layer normalization from ['layer_norm']
    axis
        The axis to normalize the
    epsilon
        The epsilon of the normalization layer
    in_channels
        Input channel

    Returns
    -------
    norm_layer
        The layer normalization layer
    """
    if isinstance(normalization, str):
        if normalization == 'layer_norm':
            assert in_channels > 0
            assert axis == -1
            norm_layer = nn.LayerNorm(normalized_shape=in_channels, eps=epsilon, **kwargs)
        elif normalization == 'rms_norm':
            assert axis == -1
            norm_layer = RMSNorm(d=in_channels, eps=epsilon, **kwargs)
        else:
            raise NotImplementedError('normalization={} is not supported'.format(normalization))
        return norm_layer
    elif normalization is None:
        return nn.Identity()
    else:
        raise NotImplementedError('The type of normalization must be str')


def _generalize_padding(x, pad_t, pad_h, pad_w, padding_type, t_pad_left=False):
    """

    Parameters
    ----------
    x
        Shape (B, T, H, W, C)
    pad_t
    pad_h
    pad_w
    padding_type
    t_pad_left

    Returns
    -------
    out
        The result after padding the x. Shape will be (B, T + pad_t, H + pad_h, W + pad_w, C)
    """
    if pad_t == 0 and pad_h == 0 and pad_w == 0:
        return x

    assert padding_type in ['zeros', 'ignore', 'nearest']
    B, T, H, W, C = x.shape

    if padding_type == 'nearest':
        return F.interpolate(x.permute(0, 4, 1, 2, 3), size=(T + pad_t, H + pad_h, W + pad_w)).permute(0, 2, 3, 4, 1)
    else:
        if t_pad_left:
            return F.pad(x, (0, 0, 0, pad_w, 0, pad_h, pad_t, 0))
        else:
            return F.pad(x, (0, 0, 0, pad_w, 0, pad_h, 0, pad_t))


def _generalize_unpadding(x, pad_t, pad_h, pad_w, padding_type):
    assert padding_type in['zeros', 'ignore', 'nearest']
    B, T, H, W, C = x.shape
    if pad_t == 0 and pad_h == 0 and pad_w == 0:
        return x

    if padding_type == 'nearest':
        return F.interpolate(x.permute(0, 4, 1, 2, 3), size=(T - pad_t, H - pad_h, W - pad_w)).permute(0, 2, 3, 4, 1)
    else:
        return x[:, :(T - pad_t), :(H - pad_h), :(W - pad_w), :].contiguous()

def apply_initialization(m,
                         linear_mode="0",
                         conv_mode="0",
                         norm_mode="0",
                         embed_mode="0"):
    if isinstance(m, nn.Linear):

        if linear_mode in ("0", ):
            nn.init.kaiming_normal_(m.weight,
                                    mode='fan_in', nonlinearity="linear")
        elif linear_mode in ("1", ):
            nn.init.kaiming_normal_(m.weight,
                                    a=0.1,
                                    mode='fan_out',
                                    nonlinearity="leaky_relu")
        else:
            raise NotImplementedError
        if hasattr(m, 'bias') and m.bias is not None:
            nn.init.zeros_(m.bias)
    elif isinstance(m, (nn.Conv2d, nn.Conv3d, nn.ConvTranspose2d, nn.ConvTranspose3d)):
        if conv_mode in ("0", ):
            nn.init.kaiming_normal_(m.weight,
                                    a=0.1,
                                    mode='fan_out',
                                    nonlinearity="leaky_relu")
        else:
            raise NotImplementedError
        if hasattr(m, 'bias') and m.bias is not None:
            nn.init.zeros_(m.bias)
    elif isinstance(m, nn.LayerNorm):
        if norm_mode in ("0", ):
            if m.elementwise_affine:
                nn.init.ones_(m.weight)
                nn.init.zeros_(m.bias)
        else:
            raise NotImplementedError
    elif isinstance(m, nn.GroupNorm):
        if norm_mode in ("0", ):
            if m.affine:
                nn.init.ones_(m.weight)
                nn.init.zeros_(m.bias)
        else:
            raise NotImplementedError
    # # pos_embed already initialized when created
    elif isinstance(m, nn.Embedding):
        if embed_mode in ("0", ):
            nn.init.trunc_normal_(m.weight.data, std=0.02)
        else:
            raise NotImplementedError
    else:
        pass



import warnings
from typing import Sequence
from shutil import copyfile
import inspect
import numpy as np
import torch
from torch import nn
from torch.nn import functional as F
import torchmetrics
import pytorch_lightning as pl
from pytorch_lightning import loggers as pl_loggers
from pytorch_lightning.callbacks import Callback
from omegaconf import OmegaConf
import os
import argparse
from einops import rearrange
from pytorch_lightning import Trainer, seed_everything
from earthformer.config import cfg
from earthformer.utils.layout import layout_to_in_out_slice
from earthformer.baselines.persistence import Persistence
from earthformer.datasets.earthnet.earthnet_dataloader import EarthNet2021LightningDataModule, get_EarthNet2021_dataloaders
from earthformer.datasets.earthnet.earthnet_scores import EarthNet2021ScoreUpdateWithoutCompute
from earthformer.datasets.earthnet.visualization import vis_earthnet_seq
from earthformer.utils.apex_ddp import ApexDDPStrategy


class PersistenceEarthNet2021PLModule(pl.LightningModule):

    def __init__(self,
                 oc_file: str = None,
                 save_dir: str = None):
        super(PersistenceEarthNet2021PLModule, self).__init__()
        if oc_file is not None:
            oc_from_file = OmegaConf.load(open(oc_file, "r"))
        else:
            oc_from_file = None
        oc = self.get_base_config(oc_from_file=oc_from_file)
        self.save_hyperparameters(oc)
        self.oc = oc
        # layout
        self.in_len = oc.layout.in_len
        self.out_len = oc.layout.out_len
        self.layout = oc.layout.layout
        self.channel_axis = self.layout.find("C")
        self.batch_axis = self.layout.find("N")
        self.channels = oc.layout.data_channels
        # logging
        self.save_dir = save_dir
        self.logging_prefix = oc.logging.logging_prefix
        # visualization
        self.train_example_data_idx_list = list(oc.vis.train_example_data_idx_list)
        self.val_example_data_idx_list = list(oc.vis.val_example_data_idx_list)
        self.test_example_data_idx_list = list(oc.vis.test_example_data_idx_list)
        self.eval_example_only = oc.vis.eval_example_only

        self.torch_nn_module = Persistence(layout=self.layout)

        test_subset_name = oc.dataset.test_subset_name
        if isinstance(test_subset_name, Sequence):
            test_subset_name = list(test_subset_name)
        else:
            test_subset_name = [test_subset_name, ]
        self.test_subset_name = test_subset_name

        self.valid_mse = torchmetrics.MeanSquaredError()
        self.valid_mae = torchmetrics.MeanAbsoluteError()
        self.valid_ens = EarthNet2021ScoreUpdateWithoutCompute(layout=self.layout, eps=1E-4)
        self.test_mse = torchmetrics.MeanSquaredError()
        self.test_mae = torchmetrics.MeanAbsoluteError()
        self.test_ens = EarthNet2021ScoreUpdateWithoutCompute(layout=self.layout, eps=1E-4)

        self.configure_save(cfg_file_path=oc_file)

    def configure_save(self, cfg_file_path=None):
        self.save_dir = os.path.join(cfg.exps_dir, self.save_dir)
        os.makedirs(self.save_dir, exist_ok=True)
        self.scores_dir = os.path.join(self.save_dir, 'scores')
        os.makedirs(self.scores_dir, exist_ok=True)
        if cfg_file_path is not None:
            cfg_file_target_path = os.path.join(self.save_dir, "cfg.yaml")
            if (not os.path.exists(cfg_file_target_path)) or \
                    (not os.path.samefile(cfg_file_path, cfg_file_target_path)):
                copyfile(cfg_file_path, cfg_file_target_path)
        self.example_save_dir = os.path.join(self.save_dir, "examples")
        os.makedirs(self.example_save_dir, exist_ok=True)

    def get_base_config(self, oc_from_file=None):
        oc = OmegaConf.create()
        oc.layout = self.get_layout_config()
        oc.optim = self.get_optim_config()
        oc.logging = self.get_logging_config()
        oc.vis = self.get_vis_config()
        if oc_from_file is not None:
            # oc = apply_omegaconf_overrides(oc, oc_from_file)
            oc = OmegaConf.merge(oc, oc_from_file)
        return oc

    @staticmethod
    def get_layout_config():
        oc = OmegaConf.create()
        oc.in_len = 10
        oc.out_len = 20
        oc.layout = "NTHWC"
        oc.data_channels = 4
        return oc

    @classmethod
    def get_dataset_config(cls):
        cfg = OmegaConf.create()
        cfg.return_mode = "minimal"
        cfg.test_subset_name = ("iid", "ood")
        layout_cfg = cls.get_layout_config()
        cfg.in_len = layout_cfg.in_len
        cfg.out_len = layout_cfg.out_len
        cfg.layout = "THWC"
        cfg.static_layout = None
        cfg.val_ratio = 0.1
        cfg.train_val_split_seed = None
        cfg.highresstatic_expand_t = False
        cfg.mesostatic_expand_t = False
        cfg.meso_crop = None
        cfg.fp16 = False        
        return cfg

    @staticmethod
    def get_optim_config():
        cfg = OmegaConf.create()
        cfg.seed = None
        cfg.micro_batch_size = 8
        return cfg

    @staticmethod
    def get_logging_config():
        cfg = OmegaConf.create()
        cfg.logging_prefix = "EarthNet2021"
        cfg.use_wandb = False
        return cfg

    @staticmethod
    def get_vis_config():
        cfg = OmegaConf.create()
        cfg.train_example_data_idx_list = [0, ]
        cfg.val_example_data_idx_list = [0, ]
        cfg.test_example_data_idx_list = [0, ]
        cfg.eval_example_only = False
        cfg.ncols = 10
        cfg.dpi = 300
        cfg.figsize = None
        cfg.font_size = 10
        cfg.y_label_rotation = 0
        cfg.y_label_offset = (-0.05, 0.4)
        return cfg

    def set_trainer_kwargs(self, **kwargs):
        r"""
        Default kwargs used when initializing pl.Trainer
        """
        callbacks = kwargs.pop("callbacks", [])
        assert isinstance(callbacks, list)
        for ele in callbacks:
            assert isinstance(ele, Callback)

        logger = kwargs.pop("logger", [])
        tb_logger = pl_loggers.TensorBoardLogger(save_dir=self.save_dir)
        csv_logger = pl_loggers.CSVLogger(save_dir=self.save_dir)
        logger += [tb_logger, csv_logger]
        if self.oc.logging.use_wandb:
            wandb_logger = pl_loggers.WandbLogger(project=self.oc.logging.logging_prefix,
                                                  save_dir=self.save_dir)
            logger += [wandb_logger, ]

        ret = dict(
            callbacks=callbacks,
            # log
            logger=logger,
            # save
            default_root_dir=self.save_dir,
            # ddp
            accelerator="gpu",
            # strategy="ddp",
            strategy=ApexDDPStrategy(find_unused_parameters=False, delay_allreduce=True),
            # optimization
            max_epochs=1,
            check_val_every_n_epoch=1,
            # NVIDIA amp
            precision=32,
        )
        return ret

    @staticmethod
    def get_earthnet2021_datamodule(
            dataset_cfg,
            micro_batch_size: int = 1,
            num_workers: int = 8):
        dm = EarthNet2021LightningDataModule(
            return_mode=dataset_cfg["return_mode"],
            val_ratio=dataset_cfg["val_ratio"],
            train_val_split_seed=dataset_cfg["train_val_split_seed"],
            layout=dataset_cfg["layout"],
            static_layout=dataset_cfg["static_layout"],
            highresstatic_expand_t=dataset_cfg["highresstatic_expand_t"],
            mesostatic_expand_t=dataset_cfg["mesostatic_expand_t"],
            meso_crop=dataset_cfg["meso_crop"],
            fp16=dataset_cfg["fp16"],
            # datamodule_only
            batch_size=micro_batch_size,
            num_workers=num_workers, )
        return dm

    @staticmethod
    def get_earthnet2021_dataloaders(
            dataset_cfg,
            micro_batch_size: int = 1,
            num_workers: int = 8):
        dataloader_dict = get_EarthNet2021_dataloaders(
            dataloader_return_mode=dataset_cfg["return_mode"],
            test_subset_name=dataset_cfg["test_subset_name"],
            val_ratio=dataset_cfg["val_ratio"],
            train_val_split_seed=dataset_cfg["train_val_split_seed"],
            layout=dataset_cfg["layout"],
            static_layout=dataset_cfg["static_layout"],
            highresstatic_expand_t=dataset_cfg["highresstatic_expand_t"],
            mesostatic_expand_t=dataset_cfg["mesostatic_expand_t"],
            meso_crop=dataset_cfg["meso_crop"],
            fp16=dataset_cfg["fp16"],
            batch_size=micro_batch_size,
            num_workers=num_workers, )
        return dataloader_dict

    @property
    def in_slice(self):
        if not hasattr(self, "_in_slice"):
            in_slice, out_slice = layout_to_in_out_slice(layout=self.layout,
                                                         in_len=self.in_len,
                                                         out_len=self.out_len)
            self._in_slice = in_slice
            self._out_slice = out_slice
        return self._in_slice

    @property
    def out_slice(self):
        if not hasattr(self, "_out_slice"):
            in_slice, out_slice = layout_to_in_out_slice(layout=self.layout,
                                                         in_len=self.in_len,
                                                         out_len=self.out_len)
            self._in_slice = in_slice
            self._out_slice = out_slice
        return self._out_slice

    def forward(self, batch):
        seq = batch[..., :self.channels]
        # mask = batch[..., self.channels:][self.out_slice].repeat_interleave(repeats=self.channels, axis=self.channel_axis)
        mask = batch[..., self.channels:][self.out_slice]
        # mask from dataloader: 1 for mask and 0 for non-masked
        in_seq = seq[self.in_slice]
        out_seq = seq[self.out_slice]
        output, _ = self.torch_nn_module(in_seq, out_seq)
        loss = F.mse_loss(output * (1 - mask), out_seq * (1 - mask))
        return output, loss

    def training_step(self, batch, batch_idx):
        seq = batch[..., :self.channels]
        mask = batch[..., self.channels:][self.out_slice]
        x = seq[self.in_slice]
        y = seq[self.out_slice]
        y_hat, loss = self(batch)
        micro_batch_size = x.shape[self.batch_axis]
        data_idx = int(batch_idx * micro_batch_size)
        if self.local_rank == 0:
            self.save_vis_step_end(
                data_idx=data_idx,
                context_seq=seq[self.in_slice].detach().float().cpu().numpy(),
                target_seq=seq[self.out_slice].detach().float().cpu().numpy(),
                pred_seq=y_hat.detach().float().cpu().numpy(),
                mode="train", )
        self.log('train_loss', loss, on_step=True, on_epoch=False)
        return loss

    def validation_step(self, batch, batch_idx, dataloader_idx=0):
        seq = batch[..., :self.channels]
        mask = batch[..., self.channels:][self.out_slice]
        x = seq[self.in_slice]
        y = seq[self.out_slice]
        micro_batch_size = x.shape[self.batch_axis]
        data_idx = int(batch_idx * micro_batch_size)
        if not self.eval_example_only or data_idx in self.val_example_data_idx_list:
            y_hat, _ = self(batch)
            if self.local_rank == 0:
                self.save_vis_step_end(
                    data_idx=data_idx,
                    context_seq=seq[self.in_slice].detach().float().cpu().numpy(),
                    target_seq=seq[self.out_slice].detach().float().cpu().numpy(),
                    pred_seq=y_hat.detach().float().cpu().numpy(),
                    mode="val", )
            if self.precision == 16:
                y_hat = y_hat.float()
            self.valid_mse(y_hat * (1 - mask), y * (1 - mask))
            self.valid_mae(y_hat * (1 - mask), y * (1 - mask))
            self.valid_ens(y_hat, y, mask)
        return None

    def validation_epoch_end(self, outputs):
        valid_mse = self.valid_mse.compute()
        valid_mae = self.valid_mae.compute()
        valid_ens_dict = self.valid_ens.compute()
        valid_loss = -valid_ens_dict["EarthNetScore"]

        self.log('valid_loss_epoch', valid_loss, prog_bar=True, on_step=False, on_epoch=True)
        self.log('valid_mse_epoch', valid_mse, prog_bar=True, on_step=False, on_epoch=True)
        self.log('valid_mae_epoch', valid_mae, prog_bar=True, on_step=False, on_epoch=True)
        self.log('valid_MAD_epoch', valid_ens_dict["MAD"], prog_bar=True, on_step=False, on_epoch=True)
        self.log('valid_OLS_epoch', valid_ens_dict["OLS"], prog_bar=True, on_step=False, on_epoch=True)
        self.log('valid_EMD_epoch', valid_ens_dict["EMD"], prog_bar=True, on_step=False, on_epoch=True)
        self.log('valid_SSIM_epoch', valid_ens_dict["SSIM"], prog_bar=True, on_step=False, on_epoch=True)
        self.log('valid_EarthNetScore_epoch', valid_ens_dict["EarthNetScore"], prog_bar=True, on_step=False,
                 on_epoch=True)
        self.valid_mse.reset()
        self.valid_mae.reset()
        self.valid_ens.reset()

    @property
    def test_epoch_count(self):
        if not hasattr(self, "_test_epoch_count"):
            self.reset_test_epoch_count()
        return self._test_epoch_count

    def increase_test_epoch_count(self, val=1):
        if not hasattr(self, "_test_epoch_count"):
            self.reset_test_epoch_count()
        self._test_epoch_count += val

    def reset_test_epoch_count(self):
        self._test_epoch_count = 0

    def test_step(self, batch, batch_idx, dataloader_idx=0):
        seq = batch[..., :self.channels]
        mask = batch[..., self.channels:][self.out_slice]
        x = seq[self.in_slice]
        y = seq[self.out_slice]
        micro_batch_size = x.shape[self.batch_axis]
        data_idx = int(batch_idx * micro_batch_size)
        if not self.eval_example_only or data_idx in self.test_example_data_idx_list:
            y_hat, _ = self(batch)
            if self.local_rank == 0:
                self.save_vis_step_end(
                    data_idx=data_idx,
                    context_seq=seq[self.in_slice].detach().float().cpu().numpy(),
                    target_seq=seq[self.out_slice].detach().float().cpu().numpy(),
                    pred_seq=y_hat.detach().float().cpu().numpy(),
                    mode="test", )
            if self.precision == 16:
                y_hat = y_hat.float()
            self.test_mse(y_hat * (1 - mask), y * (1 - mask))
            self.test_mae(y_hat * (1 - mask), y * (1 - mask))
            self.test_ens(y_hat, y, mask)
        return None

    def test_epoch_end(self, outputs):
        test_mse = self.test_mse.compute()
        test_mae = self.test_mae.compute()
        test_ens_dict = self.test_ens.compute()

        prefix = self.test_subset_name[self.test_epoch_count]
        self.log(f'{prefix}_test_mse_epoch', test_mse, prog_bar=True, on_step=False, on_epoch=True)
        self.log(f'{prefix}_test_mae_epoch', test_mae, prog_bar=True, on_step=False, on_epoch=True)
        self.log(f'{prefix}_test_MAD_epoch', test_ens_dict["MAD"], prog_bar=True, on_step=False, on_epoch=True)
        self.log(f'{prefix}_test_OLS_epoch', test_ens_dict["OLS"], prog_bar=True, on_step=False, on_epoch=True)
        self.log(f'{prefix}_test_EMD_epoch', test_ens_dict["EMD"], prog_bar=True, on_step=False, on_epoch=True)
        self.log(f'{prefix}_test_SSIM_epoch', test_ens_dict["SSIM"], prog_bar=True, on_step=False, on_epoch=True)
        self.log(f'{prefix}_test_EarthNetScore_epoch', test_ens_dict["EarthNetScore"], prog_bar=True, on_step=False,
                 on_epoch=True)
        self.test_mse.reset()
        self.test_mae.reset()
        self.test_ens.reset()

        self.increase_test_epoch_count()

    def save_vis_step_end(
            self,
            data_idx: int,
            context_seq: np.ndarray,
            target_seq: np.ndarray,
            pred_seq: np.ndarray,
            mode: str = "train", ):
        r"""
        Parameters
        ----------
        data_idx
        context_seq, target_seq, pred_seq:   np.ndarray
            layout should not include batch
        mode:   str
        """
        if self.local_rank == 0:
            if mode == "train":
                example_data_idx_list = self.train_example_data_idx_list
            elif mode == "val":
                example_data_idx_list = self.val_example_data_idx_list
            elif mode == "test":
                example_data_idx_list = self.test_example_data_idx_list
            else:
                raise ValueError(f"Wrong mode {mode}! Must be in ['train', 'val', 'test'].")
            if data_idx in example_data_idx_list:
                for variable in ["rgb", "ndvi"]:
                    save_name = f"{mode}_epoch_{self.current_epoch}_{variable}_data_{data_idx}.png"
                    vis_earthnet_seq(
                        context_np=context_seq,
                        target_np=target_seq,
                        pred_np=pred_seq,
                        ncols=self.oc.vis.ncols,
                        layout=self.layout,
                        variable=variable,
                        vegetation_mask=None,
                        cloud_mask=True,
                        save_path=os.path.join(self.example_save_dir, save_name),
                        dpi=self.oc.vis.dpi,
                        figsize=self.oc.vis.figsize,
                        font_size=self.oc.vis.font_size,
                        y_label_rotation=self.oc.vis.y_label_rotation,
                        y_label_offset=self.oc.vis.y_label_offset, )

def get_parser():
    parser = argparse.ArgumentParser()
    parser.add_argument('--save', default='tmp_earthnet2021', type=str)
    parser.add_argument('--gpus', default=1, type=int)
    parser.add_argument('--cfg', default=None, type=str)
    return parser

def main():
    parser = get_parser()
    args = parser.parse_args()
    if args.cfg is not None:
        oc_from_file = OmegaConf.load(open(args.cfg, "r"))
        dataset_cfg = OmegaConf.to_object(oc_from_file.dataset)
        micro_batch_size = oc_from_file.optim.micro_batch_size
        seed = oc_from_file.optim.seed
    else:
        dataset_cfg = OmegaConf.to_object(PersistenceEarthNet2021PLModule.get_dataset_config())
        micro_batch_size = 1
        seed = 0
    seed_everything(seed, workers=True)
    dataloader_dict = PersistenceEarthNet2021PLModule.get_earthnet2021_dataloaders(
        dataset_cfg=dataset_cfg,
        micro_batch_size=micro_batch_size,
        num_workers=8, )
    pl_module = PersistenceEarthNet2021PLModule(
        save_dir=args.save,
        oc_file=args.cfg)
    trainer_kwargs = pl_module.set_trainer_kwargs(
        devices=args.gpus,
    )
    trainer = Trainer(**trainer_kwargs)
    pl_module.reset_test_epoch_count()
    for test_dataloader in dataloader_dict["test_dataloader"]:
        trainer.test(model=pl_module,
                     dataloaders=test_dataloader)

if __name__ == "__main__":
    main()
